/*
 * ExactPatternMatchingAutomaton.cpp
 *
 *  Created on: 9. 2. 2014
 *      Author: Jan Travnicek
 */

#include "ExactPatternMatchingAutomaton.h"
#include "ExactSubtreeMatchingAutomaton.h"
#include <tree/properties/SubtreeJumpTable.h>

#include <tree/ranked/RankedTree.h>
#include <tree/ranked/RankedPattern.h>
#include <tree/ranked/PrefixRankedTree.h>
#include <tree/ranked/PrefixRankedPattern.h>
#include <tree/ranked/PrefixRankedBarTree.h>
#include <tree/ranked/PrefixRankedBarPattern.h>

#include <automaton/PDA/InputDrivenNPDA.h>
#include <automaton/PDA/VisiblyPushdownNPDA.h>
#include <automaton/PDA/NPDA.h>
#include <automaton/TA/NFTA.h>

#include <alphabet/BottomOfTheStackSymbol.h>

#include <deque>
#include <alphabet/RankedSymbol.h>

namespace arbology {

namespace exact {

automaton::Automaton ExactPatternMatchingAutomaton::construct ( const tree::Tree & pattern ) {
	return dispatch ( pattern.getData ( ) );
}

automaton::InputDrivenNPDA < > ExactPatternMatchingAutomaton::construct ( const tree::PrefixRankedTree < > & pattern ) {
	return ExactSubtreeMatchingAutomaton::construct ( pattern );
}

auto ExactPatternMatchingAutomatonPrefixRankedTree = ExactPatternMatchingAutomaton::RegistratorWrapper < automaton::InputDrivenNPDA < >, tree::PrefixRankedTree < > > ( ExactPatternMatchingAutomaton::construct );

std::vector < alphabet::Symbol > computeRHS ( const tree::PrefixRankedPattern < > & pattern, const std::vector < int > & patternSubtreeJumpTable, int i ) {
	const std::vector < std::ranked_symbol < > > & content = pattern.getContent ( );

	unsigned rank = ( unsigned ) content[i].getRank ( );

	i++;

	std::vector < alphabet::Symbol > res;

	for ( unsigned ranki = 0; ranki < rank; ranki++ ) {
		if ( content[i] == pattern.getSubtreeWildcard ( ) ) {
			res.push_back ( alphabet::Symbol ( 'R' ) );
			i++;
		} else {
			res.push_back ( alphabet::Symbol ( 'T' ) );

			i = patternSubtreeJumpTable[i];
		}
	}

	return res;
}

automaton::NPDA < > ExactPatternMatchingAutomaton::construct ( const tree::PrefixRankedPattern < > & pattern ) {
	automaton::NPDA < > res ( label::Label ( 0 ), alphabet::Symbol ( 'T' ) );

	for ( const std::ranked_symbol < > & symbol : pattern.getAlphabet ( ) ) {
		if ( symbol == pattern.getSubtreeWildcard ( ) ) continue;

		res.addInputSymbol ( alphabet::Symbol ( alphabet::RankedSymbol < > { symbol } ) );
	}

	res.setPushdownStoreAlphabet ( { alphabet::Symbol ( 'T' ), alphabet::Symbol ( 'R' ) } );

	for ( const std::ranked_symbol < > & symbol : pattern.getAlphabet ( ) ) {
		if ( symbol == pattern.getSubtreeWildcard ( ) ) continue;

		res.addTransition ( label::Label ( 0 ), alphabet::Symbol ( alphabet::RankedSymbol < > { symbol } ), std::vector < alphabet::Symbol > ( 1, alphabet::Symbol ( 'T' ) ), label::Label ( 0 ), std::vector < alphabet::Symbol > ( ( size_t ) symbol.getRank ( ), alphabet::Symbol ( 'T' ) ) );
	}

	std::vector < int > patternSubtreeJumpTable = tree::properties::SubtreeJumpTable::compute ( pattern );

	int i = 1;

	for ( const std::ranked_symbol < > & symbol : pattern.getContent ( ) ) {
		res.addState ( label::Label ( i ) );

		if ( symbol == pattern.getSubtreeWildcard ( ) )
			for ( const std::ranked_symbol < > & alphabetSymbol : pattern.getAlphabet ( ) ) {
				if ( alphabetSymbol == pattern.getSubtreeWildcard ( ) ) continue;

				if ( ( unsigned ) alphabetSymbol.getRank ( ) == 0 ) {
					res.addTransition ( label::Label ( i - 1 ), alphabet::Symbol ( alphabet::RankedSymbol < > { alphabetSymbol } ), std::vector < alphabet::Symbol > ( 1, alphabet::Symbol ( 'T' ) ), label::Label ( i - 1 ), std::vector < alphabet::Symbol > { } );

					res.addTransition ( label::Label ( i - 1 ), alphabet::Symbol ( alphabet::RankedSymbol < > { alphabetSymbol } ), std::vector < alphabet::Symbol > ( 1, alphabet::Symbol ( 'R' ) ), label::Label ( i ), std::vector < alphabet::Symbol > { } );
				} else {
					std::vector < alphabet::Symbol > push ( ( unsigned ) alphabetSymbol.getRank ( ), alphabet::Symbol ( 'T' ) );
					res.addTransition ( label::Label ( i - 1 ), alphabet::Symbol ( alphabet::RankedSymbol < > { alphabetSymbol } ), std::vector < alphabet::Symbol > ( 1, alphabet::Symbol ( 'T' ) ), label::Label ( i - 1 ), push );

					push[ ( unsigned ) alphabetSymbol.getRank ( ) - 1] = alphabet::Symbol ( 'R' );
					res.addTransition ( label::Label ( i - 1 ), alphabet::Symbol ( alphabet::RankedSymbol < > { alphabetSymbol } ), std::vector < alphabet::Symbol > ( 1, alphabet::Symbol ( 'R' ) ), label::Label ( i - 1 ), push );
				}
			}

		else
			res.addTransition ( label::Label ( i - 1 ), alphabet::Symbol ( alphabet::RankedSymbol < > { symbol } ), std::vector < alphabet::Symbol > ( 1, alphabet::Symbol ( 'T' ) ), label::Label ( i ), computeRHS ( pattern, patternSubtreeJumpTable, i - 1 ) );

		i++;
	}

	res.addFinalState ( label::Label ( i - 1 ) );
	return res;
}

auto ExactPatternMatchingAutomatonPrefixRankedPattern = ExactPatternMatchingAutomaton::RegistratorWrapper < automaton::NPDA < >, tree::PrefixRankedPattern < > > ( ExactPatternMatchingAutomaton::construct );

automaton::InputDrivenNPDA < > ExactPatternMatchingAutomaton::construct ( const tree::PrefixRankedBarTree < > & pattern ) {
	return ExactSubtreeMatchingAutomaton::construct ( pattern );
}

auto ExactPatternMatchingAutomatonPrefixRankedBarTree = ExactPatternMatchingAutomaton::RegistratorWrapper < automaton::InputDrivenNPDA < >, tree::PrefixRankedBarTree < > > ( ExactPatternMatchingAutomaton::construct );

automaton::VisiblyPushdownNPDA < > ExactPatternMatchingAutomaton::construct ( const tree::PrefixRankedBarPattern < > & pattern ) {
	automaton::VisiblyPushdownNPDA < > res ( alphabet::BottomOfTheStackSymbol::instance < alphabet::Symbol > ( ) );

	res.addState ( label::Label ( 0 ) );
	res.addInitialState ( label::Label ( 0 ) );

	for ( const std::ranked_symbol < > & symbol : pattern.getAlphabet ( ) ) {
		if ( ( symbol == pattern.getSubtreeWildcard ( ) ) || ( symbol == pattern.getVariablesBar ( ) ) ) continue;

		if ( pattern.getBars ( ).count ( symbol ) )
			res.addReturnInputSymbol ( alphabet::Symbol ( alphabet::RankedSymbol < > { symbol } ) );
		else
			res.addCallInputSymbol ( alphabet::Symbol ( alphabet::RankedSymbol < > { symbol } ) );
	}

	res.setPushdownStoreAlphabet ( { alphabet::BottomOfTheStackSymbol::instance < alphabet::Symbol > ( ) , alphabet::Symbol ( 'T' ), alphabet::Symbol ( 'R' ) } );

	for ( const std::ranked_symbol < > & symbol : pattern.getAlphabet ( ) ) {
		if ( ( symbol == pattern.getSubtreeWildcard ( ) ) || ( symbol == pattern.getVariablesBar ( ) ) ) continue;

		if ( pattern.getBars ( ).count ( symbol ) )
			res.addReturnTransition ( label::Label ( 0 ), alphabet::Symbol ( alphabet::RankedSymbol < > { symbol } ), alphabet::Symbol ( 'T' ), label::Label ( 0 ) );
		else
			res.addCallTransition ( label::Label ( 0 ), alphabet::Symbol ( alphabet::RankedSymbol < > { symbol } ), label::Label ( 0 ), alphabet::Symbol ( 'T' ) );
	}

	int i = 1;

	for ( const std::ranked_symbol < > & symbol : pattern.getContent ( ) ) {
		res.addState ( label::Label ( i ) );

		if ( symbol == pattern.getSubtreeWildcard ( ) ) {
			for ( const std::ranked_symbol < > & alphabetSymbol : pattern.getAlphabet ( ) ) {
				if ( ( alphabetSymbol == pattern.getSubtreeWildcard ( ) ) || ( alphabetSymbol == pattern.getVariablesBar ( ) ) || ( pattern.getBars ( ).count ( alphabetSymbol ) ) ) continue;

				res.addCallTransition ( label::Label ( i - 1 ), alphabet::Symbol ( alphabet::RankedSymbol < > { alphabetSymbol } ), label::Label ( i ), alphabet::Symbol ( 'R' ) );
			}
		} else if ( symbol == pattern.getVariablesBar ( ) ) {
			for ( const std::ranked_symbol < > & alphabetSymbol : pattern.getAlphabet ( ) ) {
				if ( ( alphabetSymbol == pattern.getSubtreeWildcard ( ) ) || ( alphabetSymbol == pattern.getVariablesBar ( ) ) ) continue;

				if ( pattern.getBars ( ).count ( alphabetSymbol ) )
					res.addReturnTransition ( label::Label ( i - 1 ), alphabet::Symbol ( alphabet::RankedSymbol < > { alphabetSymbol } ), alphabet::Symbol ( 'T' ), label::Label ( i - 1 ) );
				else
					res.addCallTransition ( label::Label ( i - 1 ), alphabet::Symbol ( alphabet::RankedSymbol < > { alphabetSymbol } ), label::Label ( i - 1 ), alphabet::Symbol ( 'T' ) );
			}

			for ( const std::ranked_symbol < > & alphabetSymbol : pattern.getAlphabet ( ) ) {
				if ( ( alphabetSymbol == pattern.getSubtreeWildcard ( ) ) || ( alphabetSymbol == pattern.getVariablesBar ( ) ) || ( ! pattern.getBars ( ).count ( alphabetSymbol ) ) ) continue;

				res.addReturnTransition ( label::Label ( i - 1 ), alphabet::Symbol ( alphabet::RankedSymbol < > { alphabetSymbol } ), alphabet::Symbol ( 'R' ), label::Label ( i ) );
			}
		} else if ( pattern.getBars ( ).count ( symbol ) ) {
			res.addReturnTransition ( label::Label ( i - 1 ), alphabet::Symbol ( alphabet::RankedSymbol < > { symbol } ), alphabet::Symbol ( 'T' ), label::Label ( i ) );
		} else {
			res.addCallTransition ( label::Label ( i - 1 ), alphabet::Symbol ( alphabet::RankedSymbol < > { symbol } ), label::Label ( i ), alphabet::Symbol ( 'T' ) );
		}

		i++;
	}

	res.addFinalState ( label::Label ( i - 1 ) );
	return res;
}

auto ExactPatternMatchingAutomatonPrefixRankedBarPattern = ExactPatternMatchingAutomaton::RegistratorWrapper < automaton::VisiblyPushdownNPDA < >, tree::PrefixRankedBarPattern < > > ( ExactPatternMatchingAutomaton::construct );

automaton::NFTA < > ExactPatternMatchingAutomaton::construct ( const tree::RankedTree < > & pattern ) {
	return ExactSubtreeMatchingAutomaton::construct ( pattern );
}

auto ExactPatternMatchingAutomatonRankedTree = ExactPatternMatchingAutomaton::RegistratorWrapper < automaton::NFTA < >, tree::RankedTree < > > ( ExactPatternMatchingAutomaton::construct );

label::Label constructRecursivePattern ( const std::tree < std::ranked_symbol < > > & node, automaton::NFTA < > & res, const std::ranked_symbol < > & subtreeWildcard, const label::Label & loopState, int & nextState ) {
	if ( node.getData ( ) == subtreeWildcard ) {
		label::Label state = label::Label ( nextState++ );
		res.addState ( state );

		for ( const std::ranked_symbol < > & symbol : res.getInputAlphabet ( ) ) {
			std::vector < label::Label > states;
			states.reserve ( ( size_t ) symbol.getRank ( ) );

			for ( unsigned i = 0; i < ( unsigned ) symbol.getRank ( ); i++ )
				states.push_back ( loopState );

			res.addTransition ( symbol, states, state );
		}

		return state;
	} else {
		std::vector < label::Label > states;
		states.reserve ( ( size_t ) node.getData ( ).getRank ( ) );

		for ( const std::tree < std::ranked_symbol < > > & child : node.getChildren ( ) )
			states.push_back ( constructRecursivePattern ( child, res, subtreeWildcard, loopState, nextState ) );

		label::Label state = label::Label ( nextState++ );
		res.addState ( state );
		res.addTransition ( node.getData ( ), states, state );
		return state;
	}
}

automaton::NFTA < > ExactPatternMatchingAutomaton::construct ( const tree::RankedPattern < > & pattern ) {
	std::set < std::ranked_symbol < > > alphabet = pattern.getAlphabet ( );

	alphabet.erase ( pattern.getSubtreeWildcard ( ) );

	automaton::NFTA < > res;
	res.setInputAlphabet ( alphabet );

	int nextState = 0;

	label::Label loopState = label::Label ( nextState++ );
	res.addState ( loopState );

	for ( const std::ranked_symbol < > & symbol : res.getInputAlphabet ( ) ) {
		std::vector < label::Label > states;
		states.reserve ( ( size_t ) symbol.getRank ( ) );

		for ( unsigned i = 0; i < ( unsigned ) symbol.getRank ( ); i++ )
			states.push_back ( loopState );

		res.addTransition ( symbol, states, loopState );
	}

	res.addFinalState ( constructRecursivePattern ( pattern.getContent ( ), res, pattern.getSubtreeWildcard ( ), loopState, nextState ) );
	return res;
}

auto ExactPatternMatchingAutomatonRankedPattern = ExactPatternMatchingAutomaton::RegistratorWrapper < automaton::NFTA < >, tree::RankedPattern < > > ( ExactPatternMatchingAutomaton::construct );

} /* namespace exact */

} /* namespace arbology */
