/*
 * AllEpsilonClosure.cpp
 *
 *  Created on: 12. 4. 2015
 *	  Author: Jan Travnicek
 */

#include "AllEpsilonClosure.h"

#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <set>
#include <map>
#include <iostream>

#include <regexp/properties/RegExpEpsilon.h>

namespace automaton {

namespace properties {

namespace efficient {

void process(const std::map<label::Label, std::set<label::Label>>& epsilonTransitions, const label::Label * state, std::map<label::Label, std::set<label::Label>>& closures, std::set<label::Label>& visited, std::set<label::Label>& nonvisited) {
	if(visited.count(*state)) return;

	state = &*visited.insert(*state).first;
	nonvisited.erase(*state);

	auto tos = epsilonTransitions.find(*state);
	if(tos == epsilonTransitions.end()) return;

	for(const label::Label& to : tos->second) {
		process(epsilonTransitions, &to, closures, visited, nonvisited);
		closures[*state].insert(closures[to].begin(), closures[to].end());
	}
}

std::map<label::Label, std::set<label::Label>> AllEpsilonClosure::allEpsilonClosure( const automaton::EpsilonNFA < > & fsm) {
	std::map<label::Label, std::set<label::Label>> res;

	for(const label::Label& state : fsm.getStates())
		res[state].insert(state);

	std::set<label::Label> visited;
	std::set<label::Label> nonvisited = fsm.getStates();

	std::map<label::Label, std::set<label::Label>> epsilonTransitions = fsm.getEpsilonTransitions();

	while(nonvisited.size()) {
		process(epsilonTransitions, &*nonvisited.begin(), res, visited, nonvisited);
	}

	// to correctly handle epsilon loops (swap visited and nonvisited)

	while(visited.size()) {
		process(epsilonTransitions, &*visited.begin(), res, nonvisited, visited);
	}

	return res;
}

auto AllEpsilonClosureEpsilonNFA = AllEpsilonClosure::RegistratorWrapper<std::map<label::Label, std::set<label::Label>>, automaton::EpsilonNFA < > >(AllEpsilonClosure::allEpsilonClosure);

std::map<label::Label, std::set<label::Label>> AllEpsilonClosure::allEpsilonClosure( const automaton::MultiInitialStateNFA < > & fsm) {
	std::map<label::Label, std::set<label::Label>> closure;
	for(const label::Label& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

auto AllEpsilonClosureMultiInitialStateNFA = AllEpsilonClosure::RegistratorWrapper<std::map<label::Label, std::set<label::Label>>, automaton::MultiInitialStateNFA < >>(AllEpsilonClosure::allEpsilonClosure);

std::map<label::Label, std::set<label::Label>> AllEpsilonClosure::allEpsilonClosure( const automaton::NFA < > & fsm) {
	std::map<label::Label, std::set<label::Label>> closure;
	for(const label::Label& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

auto AllEpsilonClosureNFA = AllEpsilonClosure::RegistratorWrapper<std::map<label::Label, std::set<label::Label>>, automaton::NFA < > >(AllEpsilonClosure::allEpsilonClosure);

std::map<label::Label, std::set<label::Label>> AllEpsilonClosure::allEpsilonClosure( const automaton::DFA < > & fsm) {
	std::map<label::Label, std::set<label::Label>> closure;
	for(const label::Label& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

auto AllEpsilonClosureDFA = AllEpsilonClosure::RegistratorWrapper<std::map<label::Label, std::set<label::Label>>, automaton::DFA<>>(AllEpsilonClosure::allEpsilonClosure);

std::map<label::Label, std::set<label::Label>> AllEpsilonClosure::allEpsilonClosure( const automaton::ExtendedNFA < > & fsm) {
	std::map<label::Label, std::set<label::Label>> res;

	for(const label::Label& state : fsm.getStates())
		res[state].insert(state);

	std::set<label::Label> visited;
	std::set<label::Label> nonvisited = fsm.getStates();

	std::map<label::Label, std::set<label::Label>> epsilonTransitions;
	for(const std::pair<const std::pair<label::Label, regexp::UnboundedRegExpStructure < alphabet::Symbol > >, std::set<label::Label> >& transition : fsm.getTransitions() )
		if( regexp::properties::RegExpEpsilon::languageContainsEpsilon( transition.first.second ) )
			epsilonTransitions[transition.first.first].insert(transition.second.begin(), transition.second.end());

	while(nonvisited.size()) {
		process(epsilonTransitions, &*nonvisited.begin(), res, visited, nonvisited);
	}

	// to correctly handle epsilon loops (swap visited and nonvisited)

	while(visited.size()) {
		process(epsilonTransitions, &*visited.begin(), res, nonvisited, visited);
	}

	return res;
}

auto AllEpsilonClosureExtendedNFA = AllEpsilonClosure::RegistratorWrapper<std::map<label::Label, std::set<label::Label>>, automaton::ExtendedNFA < > >(AllEpsilonClosure::allEpsilonClosure);

std::map<label::Label, std::set<label::Label>> AllEpsilonClosure::allEpsilonClosure( const automaton::CompactNFA < > & fsm) {
	std::map<label::Label, std::set<label::Label>> res;

	for(const label::Label& state : fsm.getStates())
		res[state].insert(state);

	std::set<label::Label> visited;
	std::set<label::Label> nonvisited = fsm.getStates();

	std::map<label::Label, std::set<label::Label>> epsilonTransitions;
	for(const std::pair<const std::pair<label::Label, std::vector < alphabet::Symbol > >, std::set<label::Label> >& transition : fsm.getTransitions() )
		if( transition.first.second.size() == 0 )
			epsilonTransitions[transition.first.first].insert(transition.second.begin(), transition.second.end());

	while(nonvisited.size()) {
		process(epsilonTransitions, &*nonvisited.begin(), res, visited, nonvisited);
	}

	// to correctly handle epsilon loops (swap visited and nonvisited)

	while(visited.size()) {
		process(epsilonTransitions, &*visited.begin(), res, nonvisited, visited);
	}

	return res;
}

auto AllEpsilonClosureCompactNFA = AllEpsilonClosure::RegistratorWrapper<std::map<label::Label, std::set<label::Label>>, automaton::CompactNFA < > >(AllEpsilonClosure::allEpsilonClosure);

std::map<label::Label, std::set<label::Label>> AllEpsilonClosure::allEpsilonClosure(const Automaton& automaton) {
	return dispatch(automaton.getData());
}

} /* namespace efficient */

} /* namespace properties */

} /* namespace automaton */

