/*
 * ReachableStates.cpp
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "ReachableStates.h"

#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <set>
#include <map>
#include <queue>

namespace automaton {

namespace properties {

namespace efficient {

std::set<label::Label> ReachableStates::reachableStates(const Automaton& automaton) {
	return dispatch(automaton.getData());
}

template<class T>
std::set<label::Label> ReachableStates::reachableStates( const T & fsm ) {
	std::map<label::Label, std::set<label::Label>> transitions;
	for(const auto& transition : fsm.getTransitions())
		transitions[transition.first.first].insert(transition.second.begin(), transition.second.end());

	std::deque<label::Label> queue { fsm.getInitialState( ) };
	std::set<label::Label> visited { fsm.getInitialState( ) };

	while( !queue.empty() ) {
		const std::set<label::Label>& to = transitions[queue.front()];
		queue.pop_front();

		for(const label::Label& process : to)
			if(visited.insert(process).second) {
				queue.push_back(std::move(const_cast<label::Label&>(process)));
			}
	}

	return visited;
}

auto ReachableStatesEpsilonNFA = ReachableStates::RegistratorWrapper<std::set<label::Label>, automaton::EpsilonNFA < > >(ReachableStates::reachableStates);
auto ReachableStatesNFA = ReachableStates::RegistratorWrapper<std::set<label::Label>, automaton::NFA < > >(ReachableStates::reachableStates);
auto ReachableStatesCompactNFA = ReachableStates::RegistratorWrapper<std::set<label::Label>, automaton::CompactNFA < > >(ReachableStates::reachableStates);
auto ReachableStatesExtendedNFA = ReachableStates::RegistratorWrapper<std::set<label::Label>, automaton::ExtendedNFA < > >(ReachableStates::reachableStates);

template<>
std::set<label::Label> ReachableStates::reachableStates( const automaton::MultiInitialStateNFA < > & fsm ) {
	std::map<label::Label, std::set<label::Label>> transitions;
	for(const auto& transition : fsm.getTransitions())
		transitions[transition.first.first].insert(transition.second.begin(), transition.second.end());

	std::deque<label::Label> queue ( fsm.getInitialStates( ).begin(), fsm.getInitialStates().end() );
	std::set<label::Label> visited = fsm.getInitialStates( );

	while( !queue.empty() ) {
		const std::set<label::Label>& to = transitions[queue.front()];
		queue.pop_front();

		for(const label::Label& process : to)
			if(visited.insert(process).second) {
				queue.push_back(std::move(const_cast<label::Label&>(process)));
			}
	}

	return visited;
}

auto ReachableStatesMultiInitialStateNFA = ReachableStates::RegistratorWrapper<std::set<label::Label>, automaton::MultiInitialStateNFA < >>(ReachableStates::reachableStates);

template<>
std::set<label::Label> ReachableStates::reachableStates( const automaton::DFA < > & fsm ) {
	std::map<label::Label, std::set<label::Label>> transitions;
	for(const auto& transition : fsm.getTransitions())
		transitions[transition.first.first].insert(transition.second);

	std::deque<label::Label> queue { fsm.getInitialState( ) };
	std::set<label::Label> visited { fsm.getInitialState( ) };

	while( !queue.empty() ) {
		const std::set<label::Label>& to = transitions[queue.front()];
		queue.pop_front();

		for(const label::Label& process : to)
			if(visited.insert(process).second) {
				queue.push_back(std::move(const_cast<label::Label&>(process)));
			}
	}

	return visited;
}

auto ReachableStatesDFA = ReachableStates::RegistratorWrapper<std::set<label::Label>, automaton::DFA<>>(ReachableStates::reachableStates);

} /* namespace efficient */

} /* namespace properties */

} /* namespace automaton */

